/* 
    DFT (Discrete Fourier Transform)

    Author: Kennard White

    Copyright (c) 1990-93 The Regents of the University of California.
    All rights reserved.  See the file "./copyright.h" for full copyright
    notice, limitation of liability, and disclaimer of warranty provisions.
#include "copyright.h"				COPYRIGHTENDKEY
*/
#if !defined(lint) && !defined(SABER)
static char rcsid[] = "$Header: /vol/cooley/cooley1/kennard/src/eem/RCS/dft.c,v 1.2 93/08/27 14:29:38 kennard Exp $";
#endif

#include "topFixup.h"
#include <stdio.h>
#include <math.h>

#include "topStd.h"
#include "topMem.h"
#include "topVec.h"
#include "eemDflt.h"
#include "eemCplx.h"
#include "eemDft.h"

/*global*/ char _EemDftSubPackage[] = "eem.dft";
#define SPKG _EemDftSubPackage

static int _eemDftTwidInitB = TOP_FALSE;
static TOPPtrVector _eemDftTwidTbl;


/**
    Returns an EEMComplex array of length {N}, whose contents
    are array[i] = W^(dir*i) with W = exp(-j 2 PI / N)
    and {dir} should be either +1 or -1.
**/
EEMComplex* eemDftGetTwiddles( int N, int dir) {
    EEMComplex	**ppTwid, *pTwid, *pTwidInv;
    int		i;
    if ( _eemDftTwidInitB == TOP_FALSE ) {
	topPtrVecConstructor( &_eemDftTwidTbl);
	_eemDftTwidInitB = TOP_TRUE;
    }
    topVecInsureNum( &_eemDftTwidTbl, 2*N+2);
    ppTwid = &TOPVEC_Get( &_eemDftTwidTbl, 2*N, EEMComplex*);
    if ( *ppTwid == NULL ) {
        pTwid = ppTwid[0] = (EEMComplex*) memCalloc( N, sizeof(EEMComplex));
        pTwidInv = ppTwid[1] = (EEMComplex*) memCalloc( N, sizeof(EEMComplex));
	for (i=0; i < N; i++) {
	    sincos( - M_PI*2 * i / N, &pTwid[i].i, &pTwid[i].r);
	    pTwidInv[N-i-1] = pTwid[i];
	}
    }
    return ppTwid[dir>0 ? 0 : 1];
}

/**
    Compute the DFT or IDFT of {pIn}, placing the result into {pOut}.
**/
void
eemDftCplx( int N, int dir, EEMComplex *pIn, EEMComplex *pOut) {
    int		i, j, k;
    EEMComplex	*pTwid;
    double	rsum, isum;
    double	Ninv;

    if ( N <= 0 )	return;
    pTwid = eemDftGetTwiddles( N, 1-dir);
    Ninv = 1 / N;
    for (i=0; i < N; i++) {
        k = 0;
	rsum = isum = 0.0;
	for (j=0; j < N; j++) {
	    rsum += pIn[j].r * pTwid[k].r - pIn[j].i * pTwid[k].i;
	    isum += pIn[j].r * pTwid[k].i + pIn[j].i * pTwid[k].r;
	    if ( (k -= i) < 0 )	k += N;
	}
	EEM_CPLX_SET( &pOut[i], rsum*Ninv, isum*Ninv);
    }
}

void
eemDftCplxVec( int dir, TOPVector *pIn, TOPVector *pOut) {
    int		N = TOPVEC_Num(pIn);
    topVecSetNum( pOut, N);
    eemDftCplx( N, dir, TOPVEC_Array(pIn, EEMComplex),
      TOPVEC_Array(pOut, EEMComplex));
}

EEMWindowType
eemWindowNameToType( char *wn) {
    EEMWindowType	winType;
    int			len = strlen(wn);

    if ( len < 3 )
	return EEMWinType_Null;
    /*IF*/ if ( strncasecmp( wn, "Rectangle", len)==0 ) {
	winType = EEMWinType_Rectangle;
    } else if ( strncasecmp( wn, "Bartlett", len)==0 ) {
	winType = EEMWinType_Bartlett;
    } else if ( strncasecmp( wn, "Hanning", len)==0 ) {
	winType = EEMWinType_Hanning;
    } else if ( strncasecmp( wn, "Hamming", len)==0 ) {
	winType = EEMWinType_Hamming;
    } else if ( strncasecmp( wn, "Blackman", len)==0 ) {
	winType = EEMWinType_Blackman;
    } else if ( strncasecmp( wn, "SteepBlackman", len)==0 ) {
	winType = EEMWinType_SteepBlackman;
    } else {
	return EEMWinType_Null;
    }
    return winType;
}

TOPLogical
eemWindowCoef( EEMWindowType wt, int N, double *pOut) {
    double	base_w, sin_base_w, sin_2base_w, d;
    double	scale0, scale1, scale2, freq1, freq2;
    int		i;

    if ( N < 2 )
	return TOP_FALSE;
    base_w = M_PI/(N-1);

    switch ( wt ) {
    case EEMWinType_Rectangle:
	scale0 = 1.0;
	for (i=0; i < N; i++)
	    pOut[i] = scale0;
	break;
    case EEMWinType_Bartlett:
	scale1 = 2.0/N;
	for (i=0; i < N/2; i++)
	    pOut[i] = scale1 * (i+1);
	for ( ; i < N; i++)
	    pOut[i] = scale1 * (N-(i+1));
	break;

    case EEMWinType_Hanning:
	scale0 = 0.5;  scale1 = -0.5;  freq1 = 2*base_w;
	goto hamming_proc;
    case EEMWinType_Hamming:
	scale0 = 0.54; scale1 = -0.46; freq1 = 2*base_w;
hamming_proc: ;
	freq1 = 2*base_w;
	for ( i=0; i < N; i++) {
	   pOut[i] = scale0 + scale1 * cos(freq1*i);
	}
	break;

    case EEMWinType_Blackman:
	/* This is a special case of SteepBlackman */
	d = -.16;		/* scale0 = .42, scale1 = -.5, scale2=.08 */
	goto blackman_proc;
    case EEMWinType_SteepBlackman:
	/* See Jackson 2ed, eqns 7.3.6 through 7.3.10 */
	sin_base_w = sin(base_w);
	sin_2base_w = sin(2*base_w);
	d = - TOP_SQR(sin_base_w/sin_2base_w);
blackman_proc: ;
	scale0 = (d+1)/2;	scale1 = -.5;	freq1 = 2*base_w;
				scale2 = -d/2;	freq2 = 4*base_w;
	for ( i=0; i < N; i++) {
	   pOut[i] = scale0 + scale1 * cos(freq1*i) + scale2 * cos(freq2*i);
	}
	break;

    default:
	return TOP_FALSE;
    }
    return TOP_TRUE;
}

TOPLogical
eemWindowCoefVec( EEMWindowType wt, TOPVector *pOut) {
    int		N = TOPVEC_Num(pOut);
    return eemWindowCoef( wt, N, TOPVEC_Array(pOut, double));
}
