/* 
    jef.c

    Functions for dealing with elliptical filters and jocobian
    elliptical functions.

    Author: Mei-Tjng Huang & Kennard White

    Copyright (c) 1990-93 The Regents of the University of California.
    All rights reserved.  See the file "./copyright.h" for full copyright
    notice, limitation of liability, and disclaimer of warranty provisions.
#include "copyright.h"				COPYRIGHTENDKEY
*/
#if !defined(lint) && !defined(SABER)
static char rcsid[] = "$Header: /vol/cooley/cooley1/kennard/src/eem/RCS/jef.c,v 1.16 93/08/27 14:29:36 kennard Exp $";
#endif

#include "topFixup.h"
#include <stdio.h>
#include <math.h>

#include "topStd.h"
#include "topVec.h"
#include "topMem.h"

#include "eemDflt.h"
#include "eemCplx.h"
#include "eemSilar.h"
#include "eemJef.h"

static char _SepJefSubPackage[] = "jef";
#define SPKG _SepJefSubPackage

/**
    The Jocobian Elliptical Function

    There are 12 jocobi elliptical functions (JEFs).  They are special
    cases of elliptic functions, which are doubly-periodic meromorphic
    complex functions.  A JEF is an complex function of both a complex
    variable "u" and a real parameter"m", which is the parameter of an
    elliptical integral of the first kind.  We define m' = 1 - m.  For
    computational and notational simplicity we define k, k' and alpha:
	k = sqrt(m)
	k' = sqrt(m') = sqrt(1-m) = sqrt(1-k^2)
	sin(alpha) = k = sqrt(m)
	cos(alpha) = k' = sqrt(m')
    Note that this is consistent:
	cos(alpha) = sqrt(1 - sin^2(alpha)) = sqrt(1-k^2) = k'

    When the maximum height of the elliptic filter is 1, the 
    passband ripple size (peak to peak) is 
	delta = 1 - sqrt(1/(1+ epsilon ^2)) = rho, or
	epsilon = sqrt (1/(1-d)^2 - 1)
	          sqrt (2 * delta - delta^2)/(1-delta)
    The stopband ripple size (peak to peak) is 
	1/A^2,
    where 
	k = sqrt(m) = epsilon / sqrt(A^2 -1). [Gold69]

    Also, k = w_c/w_r.  In this program, w_c is normalized to
    1; therefore, 
	k = sqrt(m) = 1/w_r and 
	w_tr = w_r - 1 = 1/k - 1 = (1-k)/k.

    The 12 JEFs can be computed from a single value phi defined by:
	phi = am(u,m)	<==> u = F(phi,m)	(am=amplitude)
        F(phi,m)=integral(0,phi) of f(theta,m) d(theta)
        f(theta,m)=1/sqrt(1-m*sin^2(theta))

    The 12 JEFs can be computed in terms of 3 "primary" JEFs:
 	sn(u,m) = sin(phi)
	cn(u,m) = cos(phi)
	dn(u,m) = sqrt(1-m*sin^2(phi)) = sqrt(1-m*sn^2(u,m))
    The JEF sn(u,m) is often refered to as *the* JEF.

    Because JEFs are doubly-perdiodic, we can talk about
    quarter-periods in the real and imaginary axis: K and iK'.  
    The points 0, K, K+iK', iK' form the vertices of a rectangle.  
    Let these four points be denoted by s, c, d, n, respectively.
    If this rectangle is translated by lK, uiK', where l and u
    are all positive and negative integers, this leads to the
    lattice:
        	.s  .c  .s  .c
		.n  .d  .n  .d
		.s  .c  .s  .c
		.n  .d  .n  .d

    This pattern is repeated indefinitely in all directions.  The
    parameters K(m) and iK'(m) are determined by the parameter m.  The
    values of K(m) and iK'(m) are important, as they tell us where the
    poles and zeros are.  K(a), for abritrary values of "a" is also of
    interest for more obscure operations.  They are found as:

	K(m) = F(pi/2,m)
	iK'(m) = iF(pi/2,m')
	K(m) = K'(m') = K'(1-m) = K'(sqrt(1-k^2))
	K'(m) = K(m') = K(1-m) = K(sqrt(1-k^2))

    It turns out that K(m) is a complete elliptical integral.
    See A&S 17.6.
    

    References
	[Gold69] Bernard Gold & Charles M. Rader, Digital Processing Of Signals,
      		 McGraw-Hill (lincoln labs) 1969, (TK 7872 F5 G6)
	[Jong82] Mark T. Jong, Methods of Disc. Sig. and Sys. Anal.,
		 McGraw-Hill 1982, (TK 7868 D5 J66)
	[IEEE79] IEEE ASSP DSP Committee, Programs For Dig. Sig. Proc., 
		 IEEE Press 1979, (TK 5102 .5 P72)
        [A & S]  M. Abramowitz and I.A. Stegun, Handbook of Mathematical 
		 Functions.  Washington, DC: National Bureau of Stanfdards, 
		 Applied MAthemiatics Series 55, 1965m (QA47.A4, Chap 16 & 17)

**/

/**
    JEF AGM : Arithmetic Geometric Mean
    Most JEF calculations are based on an AGM Table.  The table is
    intialized with a0, b0, and c0, and is iterated until numerical
    covergence is achieved.  Note that all partial values are used,
    not just the final results.  The value "m" in the structure below
    is not truely part of the AGM, but it is convientent to duplicate
    the m-value used to create AGM tables for complete elliptical integrals.

    For complete elliptical integrals K(m), use (a0,b0,c0) = (1,k',k).

    The typedefs for these types are in "eemJef.h".
**/

#define EEM_JEF_MAXAGM 200

struct _EEMJefAgm {
    double		m;
    double		a[EEM_JEF_MAXAGM];
    double		b[EEM_JEF_MAXAGM];
    double		c[EEM_JEF_MAXAGM];
    int			NumIters;
};
/**
    State for a Jacobian Elliptical Function
    All 12 JEFs are over the complex plane are determined
    by a single parameter "m".  From m we determine all other parameters.
    The CompleteAgm is an AGM table used to calculate the complete
    elliptical integral and JEF functions; the CompletePrimeAgm is
    an AGM table built with m and m_prime swapped.
**/

struct _EEMJef {
    double		m;
    double		m_prime;
    double		k;
    double		k_prime;
    double		K;
    double		K_prime;
    double              Q;
    EEMJefAgm		CompleteAGM;
    EEMJefAgm		CompletePrimeAGM;
};

struct _EEMEllip {
    EEMJef              Jef;
    EEMJef              JefK1;
    double              N;
    double              k1;
    double              u0;
    double              epsilon;
    double              Q1;
};

TOPStatus
eemJefConstructor(EEMJef *pJef, double m) {
    static char	func[] = "eemJefConstructor";

    if ( m < 0.0 || m > 1.0 ) {
        topErrMsg("%s: parameter {m} must be between 0.0 and 1.0",func);
	return -1;
    }
    eemParamsInit();
    pJef->m = m;
    pJef->m_prime = 1 - m;
    pJef->k = sqrt(m);
    pJef->k_prime = sqrt(pJef->m_prime);

    TOP_PDBG((SPKG, "For K:"))

    pJef->K       = eemJefCalcAGM( &pJef->CompleteAGM,      pJef->m, 
      1.0, pJef->k_prime, pJef->k);

    TOP_PDBG((SPKG, "For K_Prime:"))

    pJef->K_prime = eemJefCalcAGM( &pJef->CompletePrimeAGM, pJef->m_prime,
      1.0, pJef->k,       pJef->k_prime);
    pJef->Q = eemJefCalcNome( pJef->K, pJef->K_prime);

    return 0;
}

void eemJefDestructor( EEMJef *pJef) {
    pJef->m = -1;
}

EEMJef* eemJefCreate(double m) {
    EEMJef *pJef;

    pJef = MEM_ALLOC(EEMJef);
    MEM_CLEAR(pJef);
    if ( eemJefConstructor( pJef, m) < 0 ) {
	memFree(pJef);
	return NULL;
    }
    return pJef;
}


void eemJefDestroy( EEMJef *pJef) {
    eemJefDestructor( pJef);
    MEM_FREE( pJef);
}

void eemJefPrint( EEMJef *pJef, char *title) {
    printf("JEF for %s:\n", title);
    printf("\tm=%10g\tm'=%10g\n", pJef->m, pJef->m_prime);
    printf("\tk=%10g\tk'=%10g\n", pJef->k, pJef->k_prime);
    printf("\tK=%10g\tK'=%10g\n", pJef->K, pJef->K_prime);
    printf("\tQ=%10g\n\n", pJef->Q);
}

/**
    k is defined as the ratio of coplete elliptic integrals K(k)/K'(k).
    A  nome is defined as Q = exp( - M_PI *K'/K) [A&S, 17.3.17]
    The nome is an input to eemKefCalcInvNome to find k1. 
**/
double eemJefCalcNome( double K, double K_prime) {
    return exp( - M_PI * K_prime / K);
}

/** 
    Invert the nome {Q} of a JEF:
      forward:	k,m ----> K, K'  ----> Q
      reverse:	Q --> k,m
    Thus given Q, this function returns k.

    This can be used to compute the elliptical filter parameter "k1".

    Based on the fortran code ("CAY") in [IEEE79]
**/
double eemJefCalcInvNome( double Q) {
    double AD, AN, QF, QP, k;

    AN=1;
    AD=1;
    QF=1;
    QP=Q;

    /* printf("\n This is Q: %10g\n", Q); */

    do {
      QF = QF*QP;
      AN = AN+2*QF;
      if (QF <= EEMParams.Reltol)	
	break;
      QF = QF*QP;
      AD = AD+QF;
      QP = QP*Q;
      /*
       * printf("\t this is AD: %10g", AD); 
       * printf("\t this is QF: %10g\n", QF);
       */
    } while( QF > EEMParams.Reltol);

   k = 4 * sqrt(Q)*(AD*AD/(AN*AN));
   return k;
}

/**
    Calculate the Arithmetic-Geometric mean scales.
     I dont know of a physical interpretation of these scales,
    but they can be used to calculate a bunch of other things
    about the elliptical functions.
    We can easily calculate the complete elliptical integral from
    once the table is built.  Just for fun, we return this value.
    Note that this return value is only meaningful with certain values
    of (a0,b0,c0).
    Algorithm from A&S 17.6.
    Called AKK in G&M.
**/
double eemJefCalcAGM( EEMJefAgm *pAgm, double m, double a0,double b0,double c0){
int i;

    pAgm->m = m;
    pAgm->a[0] = a0;
    pAgm->b[0] = b0;
    pAgm->c[0] = c0;

    i=0;

    TOP_PDBG((SPKG, "AGM TABLE:"));
    TOP_PDBG((SPKG, "\n\ta[%d]= %g\tb[%d]=%g\tc[%d]=%g",i,pAgm->a[i],i,pAgm->b[i],i,pAgm->c[i]));
    
    for (i=1; ; i++) {
	pAgm->a[i] =     (pAgm->a[i-1] + pAgm->b[i-1])/2;
	pAgm->b[i] = sqrt(pAgm->a[i-1] * pAgm->b[i-1]);
	pAgm->c[i] =     (pAgm->a[i-1] - pAgm->b[i-1])/2;

        TOP_PDBG((SPKG, "\n\ta[%d]= %g\tb[%d]=%g\tc[%d]=%g",i,pAgm->a[i],i,pAgm->b[i],i,pAgm->c[i]));

	if ( pAgm->c[i] <= EEMParams.Reltol )
	    break;

	if ( i >= EEM_JEF_MAXAGM-1 ) {
	    topWarnMsg("AGM table did NOT converge (%g,%g,%g): %s",
	      a0, b0, c0, "results may be meaningless!!!");
	    break;
	}
    }

    TOP_PDBG((SPKG,"agm %d: (%10g %10g %10g)\n", 
      i, pAgm->a[i], pAgm->b[i], pAgm->c[i]));

    pAgm->NumIters = i;
    return( (M_PI / 2.0) / pAgm->a[pAgm->NumIters] );
}

/**
    Calculate the "phi" variable for a given real "u".
    This is the "am(u,m)" function.
    Based on [A&S, 16.4].  It uses the AGM tables.  The AGM tables
    must be initialized with {a,b,c}0 = (1,sqrt(m'),sqrt(m)).
**/

double eemJefCalcPhi( EEMJefAgm *pAgm, double u) {
    int i, N = pAgm->NumIters;
    double phi = pow(2.0, (double)N) * pAgm->a[N] * u;
    for ( i=pAgm->NumIters; i >= 1; i--) {
        phi = ( asin( sin(phi) * pAgm->c[i] / pAgm->a[i] ) + phi ) / 2;
    }
    return( phi );
}

/**
    Calculate an in-complete (partial) elliptical integral:
	u = F(phi,m)
    phi is a given angle parameter, and m is the shape parameter
    implicit in the AGM tables.  Note that K(m) = F(pi/2,m).
    This function is the inverse of eemJefCalcPhi, and can be
    used to invert JEF functions.  In particular, the
    12 JEFs are defined in terms phi with:
       phi = am(u,m)	<==>   u = F(phi,m)

    The following calculation is based on [IEEE/PDSP] function ASN
    and [A&S] 17.6.8 and 17.6.9:
    tan( phi_{n+1} - phi_{n} ) = (b_n/a_n)*tan(phi_n)
    F(phi)=phi_n/((2^n)a_n)
**/
double eemJefCalcInvPhi(EEMJefAgm *pAgm, double phi) {
    double Q;
    int i, KQ=1, KH;
    double TT, F,f1 = 0.0, two=1.0;

    /* KQ is used to keep track of the quadrant of inverse tan */

    int N=pAgm->NumIters;

    /*
     * Now, I want to do the backward calculation of eemJefCalcPhi
     * and calculate u.  It's based on A&S 17.6.8 and 17.6.9 and
     * fortran code.
     */

    for (i=1; i<=N+1; i++){
	KQ= 2 *  KQ - 1;
	Q =  pAgm->b[i-1] * tan(phi)/ pAgm->a[i-1];
	TT = atan(Q);
        f1 = phi/(pAgm->a[i-1] * two);
        two=two*2.0;
	if (TT < 0) 
	    KQ=KQ+1;
        /*  
         * I am not sure about this following  line
         */
	KH=KQ/2;  
	phi = phi + TT + M_PI * KH;
    }

      F = phi/(exp2((double)N) * pAgm->a[N]);
      /* printf ("\n this is F=%10g \t this is f1 %10g\n",F, f1); */
/*    F=phi/(exp2(N) * pAgm->a[i-1]);*/
      return (f1);
}


/**
    Calculates the primary JEFs: sn, cn, and dn.
    This is based on eemJefCalcPhi, which requires the Complete AGM
    tables. [A&S, 16.4] and [Gold69]

**/

void eemJefCalcTrigs( EEMJefAgm *pAgm, double u,
  double *sn, double *cn, double *dn) {
    double snval;
    double phi = eemJefCalcPhi( pAgm, u);
    *sn = snval = sin(phi);
    *cn = cos(phi);
    *dn = sqrt(1- pAgm->m * snval*snval );
}

/**
    Calcalates the 3 primary JEFs for a complex argment.
    This is based on eemJefCalcTrigs, which rquires that the
    AGM scales have been built properly.
    These expansions are found in [A&S, 16.21].
**/

void eemJefCalcCplxTrigs( EEMJef *pJef, EEMComplex *u, 
  EEMComplex *sn, EEMComplex *cn, EEMComplex *dn) {
    double s, c, d, s1, c1, d1, denom;
    
    eemJefCalcTrigs( &pJef->CompleteAGM, u->r, &s, &c, &d);
    eemJefCalcTrigs( &pJef->CompletePrimeAGM, u->i, &s1, &c1, &d1);
    denom = c1*c1 + pJef->m * (s*s1)*(s*s1);

    sn->r = s * d1 / denom;
    sn->i = c * d * s1 * c1 / denom;

    cn->r = c * c1 / denom;
    cn->i = - s * d * s1 * d1 / denom;

    dn->r = d * c1 * d1 / denom;
    dn->i = - pJef->m * s * c * s1 / denom;
}

/**
    Calculate inverse sn by by going backwards:
                      am         sin
        forward:  u ------> phi ------> sn
                       F          asin  
        backward: u <------ phi <------ sn
    where am and F are inverse functions.

    ******************************************************** 

    u = inverse sn(a)		a= a.r+a.i 
    a = sn(u)			u= u.r+u.i

    In this case, a is purely imaginary (a.r=0, a.i=1/epsilon).
    Because u = -j(K_{k}/N*K_{k1}) * inverse sn (j/epsilon)
    and u must be real and positive, inverse sn(j/epsilon)
    must be a imaginary number.
    
    Let s=sn(x|m), c=cn(x|m) d=dn(x|m),
        s1=sn(y|m1), c1=c(y|m1), d1=dn(y|m1)

    From [A&S,16.21.1] 
	sn(x+iy|m) = (s*d1 + j*c*d*s1*c1)/(c1^2 + m*s^2*s1^2)

    From [A&S] Table 16.5, corresponding to the argument u=0,
        sn(u) = 0, and cn(u) = dn(u) = 1.  

    After substitution:

	sn(x+iy|m) = j*s1*c1/c1^2 = j*s1/c1

    Since x=0:

	sn(iy|m)   = j*s1*c1/c1^2 = j*s1/c1

    Then, mu can be found from 
        mu = u = -K(k)*invsn(j/epsilon)/(N * K(k1));  	[Gold69]
    where invsn is found from eemJefCalcInvPhi.  
**/

double eemJefCalcU0 (EEMJef *pJef, EEMJef *pJefK1, double epsilon, int N) {
    double mu, invsn, Kk, Kk1, k, Kp, phi;
    EEMJefAgm *pAgmK, *pAgmK1;
    EEMComplex u, sn, cn, dn;

    k = pJef->k;
    Kk = pJef->K;
    Kp = pJef->K_prime;

    Kk1 = pJefK1->K;

    pAgmK  = &pJef -> CompleteAGM;
    pAgmK1 = &pJefK1-> CompleteAGM;

    pAgmK1->NumIters=pAgmK->NumIters;

    /**
     * The following line should actually read
     * inverse of sn of (j/epsilon)
    **/

    phi = atan(1.0/epsilon);
    invsn = eemJefCalcInvPhi( &pJef->CompletePrimeAGM, phi);

    mu = Kk*invsn/((double)N * Kk1);

    TOP_PDBG((SPKG,"\n for k1: N=%d, invsn=%10g\n", N, invsn));
    TOP_PDBG((SPKG,"\tu0=%10g\n", mu));

    u.r = 0;
    u.i = invsn;
   
    eemJefCalcCplxTrigs(pJef,&u, &sn, &cn, &dn);
    EEM_CPLX_Print(SPKG, &sn, "\nsn calculated from invsn should be 1/epsilon\n\t"); 
    TOP_PDBG((SPKG, "\n\n1.0/epsilon=%g", 1.0/epsilon));

    /* 
     * I don't know if I need to divide by N because the fortran code
     * didn't
     */
     return ((double)mu);
}

TOPStatus
eemEllipConstructor(EEMEllip *pEllip, int N, double m, double delta) {
    static char	func[] = "eemEllipConstructor";
    EEMJef	*pJef = &pEllip->Jef;
    EEMJef	*pJefK1 = &pEllip->JefK1;
    double	m1, epsilon;

    if ( N <= 0 ) {
        topErrMsg("%s: Parameter N must be positive (N=%d)",func,N);
	return -1;
    }
    pEllip->N = N;
    if ( eemJefConstructor(pJef, m) < 0 ) {
        topErrMsg("%s: JEF Table error (m=%g)",func,m);
	return -1;
    }
/**
    pEllip->epsilon = sqrt(2*delta -  delta*delta)/(1-delta);
**/
    pEllip->epsilon = sqrt(-1 + 1/(1-2 * delta + delta * delta));
    /*
    pEllip->epsilon = delta;
    */

    pEllip->Q1 = exp(-M_PI*(double)(pEllip->N)*(pJef->K_prime)/(pJef->K));
    epsilon = pEllip->epsilon; 
    pEllip->k1 = eemJefCalcInvNome(pEllip->Q1);
    m1 = pEllip->k1 * pEllip->k1;

    if ( eemJefConstructor(pJefK1, m1) < 0 ) {
        topErrMsg("%s: JEF Table error (m1=%g)",func,m1);
        eemJefDestructor(&pEllip->Jef);
	return -1;
    }

    pEllip->u0 = eemJefCalcU0 (pJef, pJefK1, epsilon, N);

    EEM_JEF_Print( SPKG, pJef, "k");
    EEM_JEF_Print( SPKG, pJefK1, "k1");
    return 0;
}

void eemEllipDestructor(EEMEllip *pEllip) {
    eemJefDestructor(&pEllip->Jef);
    eemJefDestructor(&pEllip->JefK1);
}

EEMEllip* eemEllipCreate(int N, double m, double delta) {
    EEMEllip	*pEllip;

    pEllip = MEM_ALLOC(EEMEllip);
    MEM_CLEAR(pEllip);
    if ( eemEllipConstructor(pEllip, N, m, delta) < 0 ) {
	MEM_FREE(pEllip);
	return NULL;
    }
    return pEllip;
}

void eemEllipDestroy(EEMEllip *pEllip) {
    eemEllipDestructor( pEllip);
    MEM_FREE(pEllip);
}

/**
    Elliptical Filters

    sn(u,m) is doubly periodic with simple poles and zeros.  In fact,
    the poles and zeros form horizontal rows, with a zero
    at the origin.  The rows are seperated by K' and columns by K.

    We define:
	T^2(u) = 1/(1 + epsilon^2 * sn^2(u,m))
    where epsilon and m are the two paramaters of the filter, determined
    by desired ripple and stop-band magnitude.

    Since we know where the singularities of sn(u,m) are, we know
    where the singularities of T^2(u) are.  It turns out that these
    are convienently placed so that we can find a rectange in the u-plane
    where one edge is surrounded by poles (the pass band), the opposite edge
    is surrounded by zeros (the stop band), and the other two edges
    are short (transition bands).

    Now if we can map this rectange in u-plane onto the jw axis of the
    s-plane in a nice way, then we can map the poles and zeros of
    T^2(u) with the same function, yielding a set of poles and zeros we
    call the elliptical filter.  Of course, we dont map every u-plane
    singularity (there are an infinite number of them); instead, we map
    the zeros on the stop band, and the poles near the pass band.  We
    must be careful to assure that the poles we map are stable.  Of
    course, there is no guarantee that a filter derived in this way
    will be useful (but obviously it is).

    The transform we use to map from u-plane to s-plane has two steps.
    The first step maps the u-plane to the l-plane (lambda-plane).
    This yields another "nice" grid pattern of poles and zeros,
    and maps the rectangle into another rectangle.  In the l-plane,
    the zeros of interest lie in a column K'(m) left of the vertical axis,
    and the poles of interest lie in a column u0 left of the vertical axis.
    The poles and zeros are seperated by 2K(m)/N vertically.
    The locations of these poles and zeros in the l-plane is particularly
    important, because this is where most programs start.

    The second stepmaps the l-plane to the s-plane.  Amazingly, this 
    transformation is:
	f:l->s = j * sn(-j*l,m)
    with certain constraints on some parameters.
    This complex sn() can be defined in terms of a real sn().

    Summary
    To design an elliptical filter:
    1. Convert your particular design parameters into 
        (A) the order N,
        (b) the cutoff frequency Wc,
        (c) the JEF-parameter m (often via m=k^2=(wC/Wr)^2, Wr=stop freq)
	(d) the passband ripple parameter e (eplison)

    2.  Calculate the u0, K(m), and K'(m) parameters.  Generate the
	the l-plane poles and zeros based on these values.
    3.  Transform these poles and zeros into the s-plane.


    More details:
    To transform from the l plane to the s plane, one makes the conformal
	    transformation s=j(wc)sn(-jl,k).

    If s=s.r+js.i and l=q+jr, then

        s.r=wc[sn(q.k')*cn(q,k')*cn(r,k)*dn(r,k)/[1-sq(sn(q,k'))*sq(dn(r,k))]
      
    and s.i = (wc)[sn(r,k)*dn(q,k') / [1 - sq(sn(q,k')) * sq(dn(r,k)) ]

    The zeros all lie on the jw axis in the S-plane.

    There are a total of 2*N singularities.  However, if N is odd, one pair
    of pole and zero will lie on the q axis in the l-plane.  Hence, when
    transformed into the S-plane, the zero will transform into infinity on
    the jw axis in the S-plane; so this zero is not computed.  The pole is
    computed separately because this pole does not have a conjugate pair.

    Note that if N is even, the poles and zeros always appear in conjugate
    pairs, respectively.
**/

void eemJefEllipSilars( EEMEllip *pEllip, double w_pass, EEMSilarGrp *pSilarGrp) {
    int		i;
    int		N = pEllip->N;
    EEMJef	*pJef = &pEllip->Jef;
    double	epsilon = pEllip->epsilon;
    int		NZ = N/2;
    double	H_Sq;
    EEMComplex	l, ll, sn, cn, dn, z, p, zero, H_zero;

    EEM_SILAR_Gain(pSilarGrp).r = 1;
    EEM_SILAR_Gain(pSilarGrp).i = 0;

    /*
     * Zero locations in the S plane
     */
    for ( i=1; i <= NZ; i++) {
        l.r = - pJef->K_prime;
	l.i = ((N+1-2*i) / (double)N) * pJef->K;

	ll.r = l.i;
	ll.i = -l.r;

	eemJefCalcCplxTrigs( pJef, &ll, &sn, &cn, &dn); 
 
	z.r = -sn.i * w_pass;
	z.i = sn.r * w_pass;

	eemSilarGrpAppend( pSilarGrp, EEMSilarF_Conj, 0.0, z.i, 1);
    }

    /*
     * Pole locations in the S plane
    */
    for ( i=1; i <= NZ; i++) {
        l.r = - pEllip->u0;
        l.i = ((N+1-2*i) / (double)N) * pJef->K;

	ll.r = l.i;
	ll.i = -l.r;

        eemJefCalcCplxTrigs( pJef, &ll, &sn, &cn, &dn);

        p.r = - sn.i * w_pass;
	p.i =   sn.r * w_pass;
	
	eemSilarGrpAppend( pSilarGrp, EEMSilarF_Conj, p.r, p.i, -1);
    }

    if (N & 1) {
        /* 
         *There is another pole on the q axis in the l plane if N=odd
         */
        l.r = - pEllip->u0;
	l.i = 0;

        ll.r = l.i;
        ll.i = -l.r;

        eemJefCalcCplxTrigs( pJef, &ll, &sn, &cn, &dn);

        p.r = - sn.i * w_pass;
        p.i =   sn.r * w_pass;

	eemSilarGrpAppend( pSilarGrp, 0, p.r, p.i, -1);
    }
 
    zero.r = 0.0; 
    zero.i = 0.0;
    eemSilarGrpEval( pSilarGrp, &zero, &H_zero); 
    H_Sq = EEM_CPLX_MAGSQ( &H_zero);  

    if (N & 1) 
	pSilarGrp->gain.r = 1.0/sqrt(H_Sq);
	/* topCplxInv2( &pSilarGrp->gain, &H_zero); */
    else 
        pSilarGrp->gain.r = 1.0/sqrt(H_Sq * (1 + epsilon * epsilon));
    pSilarGrp->gain.i = 0.0;
}

TOPStatus
eemEllipFilter( EEMSilarGrp *pSilarGrp, int N, double w_pass,
  double m, double delta){
    static char	func[] = "eemEllipFilter";
    EEMEllip ellip; 

    if ( eemEllipConstructor( &ellip, N, m, delta) < 0 ) {
	topErrMsg("%s: EEMEllip construction failed", func); 
	return -1;
    }
    eemJefEllipSilars( &ellip, w_pass, pSilarGrp);
    eemEllipDestructor( &ellip);
    return 0;
}
