# Create a continuous-time butterworth filter
# Parameter {N} MUST be a manifest integer constant.
#

proc filters.tcl { } { }

proc ctButterworth { fltr N w_pass } {
    global PI
    set tgtMod [view tag $fltr]
    set w_pass [expr $w_pass]
    set deltaphi [expr { 2*$PI / (2*$N) }]
    set ofsphi [expr { $PI/2 + $deltaphi/2 }]
    set h_gain 1.0
    for {set i 0} {$i<($N/2)} {incr i} {
	set phi [expr {$ofsphi + $i * $deltaphi}]
	$tgtMod insert "mp $w_pass $phi" -1 conj
	set h_gain [expr {$h_gain * $w_pass * $w_pass}]
    }
    if { $N & 1 } then {
	$tgtMod insert [expr { - $w_pass}] -1
	set h_gain [expr {$h_gain * $w_pass}]
    }
    $tgtMod gain $h_gain
}

proc silarMag { r i } {
    return [expr [dflt sqrt {$r*$r + $i*$i}]]
}

proc ctChebyshev1 { fltr N w_pass delta } {
    global PI
    set tgtMod [view tag $fltr]
    set w_pass [expr $w_pass]
    set delta [expr $delta]
    if { $delta <= 0.0 || $delta >= 1.0 } {
	error "ctChebyshev1: delta must be between 0.0 and 1.0"
    }
    set eps [expr [dflt sqrt {1.0/((1-$delta)*(1-$delta)) - 1.0}]]
    set gammaP [expr { (1.0 + [expr [dflt sqrt {1+$eps*$eps}]])/$eps }]
    set gamma [expr [dflt pow {$gammaP} {1.0/$N}]]
    set wc_sinh_phi [expr {$w_pass*($gamma - 1.0/$gamma)/2}]
    set wc_cosh_phi [expr {$w_pass*($gamma + 1.0/$gamma)/2}]
    set h_gain 1.0
    for {set k 0} {$k<($N/2)} {incr k} {
        set u_k [expr { (2*($k+1) - 1)*$PI/(2*$N) }]
	set pole_r [expr { - $wc_sinh_phi * [expr [dflt sin $u_k]] }]
	set pole_i [expr {   $wc_cosh_phi * [expr [dflt cos $u_k]] }]
	$tgtMod insert "$pole_r $pole_i" -1 conj
	set pole_mag [silarMag $pole_r $pole_i]
	set h_gain [expr {$h_gain * $pole_mag * $pole_mag}]
    }
    if { $N & 1 } then {
	$tgtMod insert [expr {- $wc_sinh_phi}] -1
	set h_gain [expr {$h_gain * $wc_sinh_phi} ]
    }
    $tgtMod gain $h_gain
}

#
# Smooth in passband, ripple in stopband
#
proc ctChebyshev2 { fltr N w_stop delta } {
    global PI
    set tgtMod [view tag $fltr]
    set delta [expr $delta]
    if { $delta <= 0.0 || $delta >= 1.0 } {
	error "ctChebyshev2: delta must be between 0.0 and 1.0"
    }
    set gammaP [expr { (1.0 + [dflt sqrt {1.0-$delta*$delta}])/$delta }]
    set gamma [dflt pow {$gammaP}  {1.0/$N}]
    set wc_sinh_phi [expr { ($gamma - 1.0/$gamma)/2.0 } ]
    set wc_cosh_phi [expr { ($gamma + 1.0/$gamma)/2.0 } ]
    set h_gain 1.0
    for {set k 0} {$k<($N/2)} {incr k} {
        set u_k [expr { (2.0*($k+1) - 1.0)*$PI/(2.0*$N) }]
	set p_r [expr { - $wc_sinh_phi * [dflt sin $u_k] }]
	set p_i [expr {   $wc_cosh_phi * [dflt cos $u_k] }]
	set pole_denom [expr {$p_r*$p_r + $p_i*$p_i}]
	set pole_r [expr {  $w_stop*$p_r/$pole_denom}]
	set pole_i [expr {- $w_stop*$p_i/$pole_denom}]
	$tgtMod insert "$pole_r $pole_i" -1 conj
	set pole_mag [silarMag $pole_r $pole_i]
	set h_gain [expr {$h_gain * $pole_mag * $pole_mag}]
	set zero_i [expr {$w_stop/[dflt cos $u_k]} ]
	$tgtMod insert "0 $zero_i" 1 conj
	set h_gain [expr {$h_gain / ($zero_i * $zero_i)}]
    }
    if { $N & 1 } then {
	set pole_r [expr { - $w_stop / $wc_sinh_phi }]
	$tgtMod insert "$pole_r" -1
	set h_gain [expr {$h_gain * (- $pole_r)}]
    }
    $tgtMod gain $h_gain
}

proc ctChebyshev2_old { fltr N w_pass w_tr delta } {
    global PI
    set w_pass [expr $w_pass]
    set w_stop [expr {$w_pass+$w_tr}]
    return [ctChebyshev2_core $fltr $N $w_stop $delta]
}

proc ctElliptical { fltr N w_pass w_tr delta } {
    set tgtMod [view tag $fltr]
    set w_pass [expr $w_pass]
    set w_stop [expr {$w_pass+$w_tr}]
    set delta [expr $delta]
    if { $delta <= 0.0 || $delta >= 1.0 } {
	error "ctElliptical: delta must be between 0.0 and 1.0"
    }
    set k [expr {$w_pass/($w_stop)}]
    set m [expr {$k*$k}]
    set silargroup [ellip $N $w_pass $m $delta]
    foreach silar [lrange $silargroup 1 end] {
#puts stdout "silar is $silar"
	set sil_loc [lindex $silar 0]
	set sil_order [lindex $silar 1]
	set sil_flags [lindex $silar 2]
	$tgtMod insert $sil_loc $sil_order $sil_flags
    }
    $tgtMod gain [lindex $silargroup 0]
}

###############################################################################
# A SpectralXform pipe requires extra state to describe the
# specific type of transformation.  These values are stored
# in a TCL array variable $pipeName.  They are:
#	Type	Which of the 8 possible transforms
#	OldLo	Lo & Hi cutoffs of the "source" filter
#	OldHi
#	NewLo	Lo & Hi cutoffs of the "target" filter
#	NewHi	
#	NumFac
#	T	Sampling time (for Z transformations only)
#	Alpha	Derivived parameters for transform
#	Beta
#	K
#
# For slptobp, some Jackson uses p = (s^2 + lo*hi)/(s(hi-lo)),
# while Parks uses p = (s^2 + lo*hi)/s.  The advantage of Jackson's
# method is that the width of the passband doesn't change, while Parks
# requires an adjustment wp = (hi^2 - lo*hi)/hi.
###############################################################################

proc xpSpectralXform { name srcMod tgtMod } {
    global xpole
    if ![info exists xpole(SpectralXform)] {
	xppipeclass SpectralXform SilarRep SilarRep
	set xpole(SpectralXform) 1
    }
    set name [view pipe $srcMod SpectralXform $tgtMod $name]
    upvar #0 $name pvars

    set pvars(active) 0
    set pvars(config1) 1
    set pvars(config2) [info exists pvars(NewWidth)]
    set pvars(T) 1.0

    trace var pvars(Type)  w "_xpPSX.TraceConfig1CB $name"
    trace var pvars(OldLo) w "_xpPSX.TraceConfig1CB $name"
    trace var pvars(OldHi) w "_xpPSX.TraceConfig1CB $name"
    trace var pvars(NewLo) w "_xpPSX.TraceConfig1CB $name"
    trace var pvars(NewHi) w "_xpPSX.TraceConfig1CB $name"

    trace var pvars(OldCenter) w "_xpPSX.TraceConfig2CB $name"
    trace var pvars(OldWidth)  w "_xpPSX.TraceConfig2CB $name"
    trace var pvars(NewCenter) w "_xpPSX.TraceConfig2CB $name"
    trace var pvars(NewWidth)  w "_xpPSX.TraceConfig2CB $name"

    # Trigger the trace now to recalc everything
    if [info exist pvars(Type)] {
        set pvars(active) 1
    }
    return $name
}

proc _xpPSX.TraceConfig1CB { pipe n1 n2 op } {
    upvar #0 $pipe pvars 

    if { ! $pvars(active) } { set pvars(config1) 1; return }

    set OldLo $pvars(OldLo)
    set OldHi $pvars(OldHi)
    set NewLo $pvars(NewLo)
    set NewHi $pvars(NewHi)
    set T     $pvars(T)

    set pvars(NumFac) 1
    case $pvars(Type) {
	slptolp {
	    set pvars(Op)	mul
    	    set pvars(Alpha)	[expr {$NewHi/$OldHi}]
	}
	slptohp {
	    set pvars(Op)	div
    	    set pvars(Alpha)	[expr {$NewHi*$OldHi}]
	    set pvars(InfLoc)	0
	}
	slptobp {
	    set pvars(Op)	mul
    	    set pvars(Beta)	[expr {($NewLo-$NewHi)/$OldHi}]
    	    set pvars(Gamma)	[expr {($NewLo*$NewHi)}]
    	    set pvars(NumFac)	2
	    set pvars(InfLoc)	0
	}
	slptobs {
	    set pvars(Op)	div
    	    set pvars(Beta)	[expr {($NewLo-$NewHi)*$OldHi}]
    	    set pvars(Gamma)	[expr {($NewLo*$NewHi)}]
    	    set pvars(NumFac)	2
	    set pvars(InfLoc)	"[cplx sqrt [expr {-$NewLo*$NewHi}]]"
	}
	zlptolp {
	    set pvars(Op)	add
    	    set topalpha	[dflt sin {$T*($OldHi - $NewHi)/2.0}]
    	    set botalpha	[dflt sin {$T*($OldHi + $NewHi)/2.0}]
    	    set pvars(Alpha)	[expr { $topalpha / ($botalpha+0.0) }]
	}
	zlptohp {
	    set pvars(Op)	sub
    	    set topalpha	[dflt cos {$T*($NewHi + $OldHi)/2.0}]
    	    set botalpha	[dflt cos {$T*($NewHi - $OldHi)/2.0}]
    	    set pvars(Alpha)	[expr { $topalpha / ($botalpha+0.0) }]
	}
	{zlptobp zlptobs} { 
    	    set topalpha [dflt cos {$T*($NewHi + $NewLo)/2.0}]
    	    set botalpha [dflt cos {$T*($NewHi - $NewLo)/2.0}]
    	    set alpha [expr { $topalpha / ($botalpha+0.0) }]
    	    set k1 [dflt tan {$T*(         $OldHi)/2.0}]
    	    set k2 [dflt tan {$T*($NewHi - $NewLo)/2.0}]
	    if { "$pvars(Type)"=="zlptobp" } {
		set K [expr { $k1 / $k2 }]
		set pvars(Beta1) [expr { -2.0 * $alpha * $K/($K+1.0) }]
		set pvars(Beta2) [expr { ($K-1.0)/($K+1.0) }]
		set pvars(Op)	add
	    } else {
		set K [expr { $k1 * $k2 }]
		set pvars(Beta1) [expr { -2.0 * $alpha/($K+1.0) }]
		set pvars(Beta2) [expr { (1.0-$K)/(1.0+$K) }]
		set pvars(Op)	sub
	    }
    	    set pvars(NumFac) 2
#puts stdout "zlpto alpha $alpha K $K beta $pvars(Beta1) $pvars(Beta2)"
	}
    }
    set pvars(config1) 0
}

#
# Map from (center & width) to (lo & hi)
#
proc _xpPSX.TraceConfig2CB { pipe n1 n2 op } {
    upvar #0 $pipe pvars 

    if { ! $pvars(active) } { set pvars(config2) 1; return }
    set pvars(active) 0

    set pvars(OldLo) [expr $pvars(OldCenter)-$pvars(OldWidth)]
    set pvars(OldHi) [expr $pvars(OldCenter)+$pvars(OldWidth)]
    set pvars(NewLo) [expr $pvars(NewCenter)-$pvars(NewWidth)]
    set pvars(NewHi) [expr $pvars(NewCenter)+$pvars(NewWidth)]

    set pvars(config2) 0
    set pvars(active) 1
}

proc _xpPipeSpectralXformNotify { pipe srcMod tgtMod idxList netPower } {
    upvar #0 $pipe pvars

    if $pvars(config2)		{ _xpPSX.TraceConfig2CB $pipe xn1 xn2 xop }
    if $pvars(config1)		{ _xpPSX.TraceConfig1CB $pipe xn1 xn2 xop }

    set type $pvars(Type)
    set op $pvars(Op)
    $tgtMod gain [$srcMod gain]
    set srcNum [$srcMod num]
    $tgtMod num [expr {$srcNum*$pvars(NumFac)}]

    foreach idx $idxList {
	set src [$srcMod silar $idx]
	set srcLoc [lindex $src 0]
	set srcPow [lindex $src 1]
	set srcFlags [lrange $src 2 end]
	case $type in {
	    {slptolp slptohp} {
		set tgtLoc [cplx $op $pvars(Alpha) $srcLoc]
		$tgtMod silar !$idx $tgtLoc $srcPow @$srcFlags
	    }
	    {slptobp slptobs} {
		set tgtLocs [cplx quadroots 1 \
		  [cplx $op $pvars(Beta) $srcLoc] $pvars(Gamma) ]
		$tgtMod silar ![expr 2*$idx+0] \
		  [lindex $tgtLocs 0] $srcPow @$srcFlags
		$tgtMod silar ![expr 2*$idx+1] \
		  [lindex $tgtLocs 1] $srcPow @$srcFlags
	    }
	    {zlptolp zlptohp} {
		set tgtLoc [cplx div \
		  [cplx $op $pvars(Alpha) $srcLoc] \
		  [cplx $op 1.0 [cplx mul $pvars(Alpha) $srcLoc]] ]
		$tgtMod silar !$idx $tgtLoc $srcPow @$srcFlags
	    }
	    {zlptobp zlptobs} {
# puts stdout "silar idx $idx src ``$src'' flags $srcFlags"
#puts stdout "silar idx $idx $srcLoc"
		set Beta1 $pvars(Beta1)
		set Beta2 $pvars(Beta2)
		set tgtLocs [cplx quadroots \
		  [cplx $op [cplx mul $srcLoc $Beta2] 1.0] \
		  [cplx mul $Beta1 [cplx $op $srcLoc 1.0]] \
		  [cplx $op $srcLoc $Beta2] ]
		$tgtMod silar ![expr 2*$idx+0] \
		  [lindex $tgtLocs 0] $srcPow @$srcFlags
		$tgtMod silar ![expr 2*$idx+1] \
		  [lindex $tgtLocs 1] $srcPow @$srcFlags
	    }
	    {default} {
		error [format "%s: Unknown SpectralXform type %s\n" \
		  $pvars $type]
	    }
	}
    }


    #
    # Take care of silars at infinity
    #
    set idx [expr {$srcNum * $pvars(NumFac)}]
    case $type in {
	{slptohp slptobp} { 
	    $tgtMod silar !$idx $pvars(InfLoc) [expr {- $netPower}] @
	}
	{slptobs} {
	    $tgtMod silar !$idx $pvars(InfLoc) [expr {- $netPower}] @{conj}
	}
	default {
	    $tgtMod silar !$idx 0 0 @
	}
    }
}

proc xpSilarMapping { mapType name srcMod tgtMod } {
    upvar #0 $name mvars
    set mvars(maptype) $mapType
    set name [view pipe $srcMod PolezeroMap $tgtMod $name]
}


###############################################################################
# 			
###############################################################################

proc xpFourierDesign { filter } {
    filter create $filter z
    set fsampView $filter.fsamp
    view get $fsampView CplxSeqFreq
    set impulseView $filter.fsamp-impulse
    set fourierView $filter.fsamp-impulse-fourier
    set fsampPipe foo
    set fftPipe goo
    set fsampPipe [view pipe $fsampView FsampImpulse $impulseView $fsampPipe]
    set fftPipe [view pipe $impulseView ImpulseFourier $fourierView $fftPipe]

    xpView.Display $fsampView
}

proc xpBilinearPair { h1 h2 } {
    filter create $h1 s
    view get $h1.polezero SilarRep
    xpView.Display $h1.polezero

    filter create $h2 z
    view get $h2.polezero SilarRep
    xpView.Display $h2.polezero
    
    set mapPipe foo
    upvar #0 $mapPipe mvars
    set mvars(maptype) bilinear
    view pipe $h1.polezero PolezeroMap:B $h2.polezero $mapPipe
}


proc xpDftPair { h1 h2 } {
    filter create $h1 z
    view get $h1.impulse CplxSeqTime
    view index $h1.impulse 16 0.0 1.0
    filter create $h2 z
    view get $h2.fourier CplxSeqFreq

    set mapPipe hoo
    upvar #0 $mapPipe mvars
#    set mvars(mapping) bilinear
    view pipe $h1.impulse ImpulseFourier $h2.fourier $mapPipe

    xpView.Display $h1.impulse
    xpView.Display $h2.fourier
    
}
