#
#	stdfltrdlg.tcl :: Standard Filter Dialog UI
#
#	Parameters:
#	  order		order of the filter
#	  w_width	width of band (rad/sec, single side)
#			(pass-band width for lp&bp, stop-band width for bs&hp)
#	  w_tr		transition band width (red/sec)
#	  w_center	center of filter, for bp & bs (0 for lp & hp)
#			(center of pass-band for lp&bp, stop-band for bs&hp)

#
#	Depending on the current filter selection made by the user,
#	we need to make windows appear and disappear.  These
#	function do that; they are activated by TCL traces.
#

proc stdfltrdlg.tcl { } { }

proc xpSFD.Config { dlg } {
    upvar #0 $dlg fvars

    case $fvars(design) in {
        butterworth {
	    set cmd ctButterworth; set params "order w_pass"
        }
        chebyshev1 {
	    set cmd ctChebyshev1; set params "order w_pass delta"
        }
        chebyshev2 {
	    set cmd ctChebyshev2; set params "order w_pass delta"
        }
        elliptical {
	    set cmd ctElliptical; set params "order w_pass w_tr delta"
        }
    }
    set fvars(cmdparams) $params

    case $fvars(pass) in {
	{bp bs} {
	    lappend params "w_center"
	}
    }
    case $fvars(mapping) in {
      impulse {
	lappend params "n_zeros"
      }
    }

    if { "$fvars(domain)"=="z" } {
	set dd $dlg.mid.pass.banddomain
	if { "$fvars(pass)"!="lp" } {
	    $dd.title conf -fore [lindex [$dd.s conf -fore] 4]
	    foreach p { s z } {
	        if { "[lindex [$dd.$p conf -state] 4]"=="disabled" } {
	           $dd.$p conf -state normal
		}
	    }
	} else {
	    $dd.title conf -fore [lindex [$dd.s conf -disabledfore] 4]
	    foreach p { s z } {
		$dd.$p conf -state disabled
	    }
	}

	set dd $dlg.mid.map; set p prewarp
	if { "$fvars(mapping)"=="bilinear" } {
	    if { "[lindex [$dd.$p conf -state] 4]"=="disabled" } {
	       $dd.$p conf -state normal
	    }
	} else {
	    $dd.$p conf -state disabled
	}
    }

    set fvars(cmd) $cmd
    set fvars(params) $params

    if { $fvars(autoupdate) } {
	xpSFD.SafeUpdate $dlg
    }
}

# : ( "$fvars(passdomain)"=="$p" ? "active" : "normal" ) } ]

proc xpSFD.TraceDesign { dlg n1 n2 op } {
    upvar #0 $dlg fvars
    set fvars(needUpdateB)	1
    if [catch "xpSFD.Config $dlg" error] {
	tkerror "Config failed in xpSFD.TraceDesign"
    }
}

proc xpSFD.TraceParams { dlg n1 n2 op } {
    upvar #0 $dlg fvars

    set oldparams  $fvars(oldparams)
    set newparams  $fvars(params)
    if { [string compare $oldparams $newparams]==0 } {
	return
    }
    foreach p $oldparams {
	pack unpack $dlg.mid.vals.$p
    }
    foreach p $newparams {
	pack append $dlg.mid.vals $dlg.mid.vals.$p { top fillx }
    }
    set fvars(oldparams) $newparams
}

proc xpSFD.TraceVals { dlg n1 n2 op } {
    upvar #0 $dlg fvars

    if $fvars(updateactive) {
	return
    }
    set p [lindex [split $n2 -] 1]
    if { $fvars(autoupdate) } {
	xpSFD.SafeUpdateParams $dlg
    }
}

proc xpSFD.TraceUpdate { dlg n1 n2 op } {
    upvar #0 $dlg fvars
    if { $fvars(autoupdate) } {
	xpSFD.SafeUpdateParams $dlg
    }
}

proc xpSFD.ValCB { dlg valname val } {
    upvar #0 $dlg fvars

    set fvars(val-$valname) $val
    return $val
}

proc xpSFD.GetParams { dlg } {
    upvar #0 $dlg fvars

    set fvars(updateactive) 1
    foreach p $fvars(params) {
        set val [$dlg.mid.vals.$p.val get]
        set fvars(val-$p) $val
    }
    set fvars(updateactive) 0
}

#
# There are 3 possible pipes in existence:
#   sz	Silar mapping (bilinear) from s-domain to z-domain
#   sx	Spectral Transformation from s-lp to s-pass
#   zx	Spectral Transformation from z-lp to z-pass
#
# This results in 5 possible configurations:
# 	s.tgt
#	s.tgt-lp    --sx-->  s.tgt
#	s.tgt-s-lp  --sx-->  s.tgt-s   --sz-->  z.tgt
#	s.tgt-lp-s  --sz-->  z.tgt-lp  --zx-->  z.tgt
#	s.tgt-s     --sz-->  z.tgt
#
# However, only one of {sx,zx} is in existance at any given time,
# so we collapse the two together.
#
proc xpSFD.UpdatePipes { dlg } {
    global errorInfo
    set func xpSFD.Update
    upvar #0 $dlg fvars $dlg.sx sx $dlg.sz sz

    set viewspec $fvars(viewspec)
    set viewinfo [view get $viewspec SilarRep]
    set viewfltr [lindex $viewinfo 0]
    set viewname [lindex $viewinfo 1]
    set viewpath $viewfltr.$viewname

    set dstpath $viewpath

    if [info exist sx] { unset sx }
    if [info exist sz] { unset sz }

    set activate ""
    set allviews $dstpath

    # 
    # For z-domain filters, create the fake s-filter from which
    # the z-filter is dirived.
    #
    if { "$fvars(domain)"=="z" } {
	set sfltrinfo [lindex [filter get $viewfltr-s(s)] 0]
	set sfltr [lindex $sfltrinfo 0]
    }

    # 
    # Setup s-->z mapping (post spectral transformation (sx in s-domain))
    #
    if { "$fvars(domain)"=="z" && "$fvars(passdomain)"=="s" } {
	set srcinfo [xpView.GetByClass $sfltr.$viewname-s SilarRep]
	set srcpath [lindex $srcinfo 0].[lindex $srcinfo 1]
	xpSilarMapping $fvars(mapping) $dlg.sz $srcpath $dstpath
	set dstpath $srcpath
	set activate "sz $activate"
        set allviews "$dstpath $allviews"
    }

    #
    # Setup s or z-domain spectral transformation (or not)
    #
    if { "$fvars(pass)"!="lp" } {
	set srcinfo [view get $dstpath-lp SilarRep]
	set srcpath [lindex $srcinfo 0].[lindex $srcinfo 1]
	set srcfltrinfo [lindex [filter get [lindex $srcinfo 0]] 0]
	set srcdomain [lindex $srcfltrinfo 1]
	xpSpectralXform $dlg.sx $srcpath $dstpath
	set dstpath $srcpath
	set activate "sx $activate"
        set allviews "$dstpath $allviews"
    }

    # 
    # Setup s-->z mapping (pre spectral transformation (sx in z domain))
    # 
    if { "$fvars(domain)"=="z" && "$fvars(passdomain)"=="z" } {
	set srcinfo [xpView.GetByClass $sfltr.$viewname-s SilarRep]
	set srcpath [lindex $srcinfo 0].[lindex $srcinfo 1]
	xpSilarMapping $fvars(mapping) $dlg.sz $srcpath $dstpath
	set dstpath $srcpath
	set activate "sz $activate"
        set allviews "$dstpath $allviews"
    }

    set fvars(slp) $dstpath
    set fvars(activate) $activate
    set fvars(allviews) $allviews
}

proc xpSFD.PreWarpVal { dlg val } {
    upvar #0 $dlg fvars
    if { "$fvars(domain)"=="z" && "$fvars(mapping)"=="bilinear" } {
        set T 1.0
        return [expr { [dflt tan {$val * $T/2.0}] * 2.0 / $T } ]
    }
    return $val
}

proc xpSFD.PreWarpRange { dlg center val } {
    return [expr { 
        ( [xpSFD.PreWarpVal $dlg [expr {$center+$val}]]
      - [xpSFD.PreWarpVal $dlg [expr {$center-$val}]] ) / 2.0 } ]
}

proc xpSFD.PostWarpVal { dlg val } {
    upvar #0 $dlg fvars
    if { "$fvars(domain)"=="z" && "$fvars(mapping)"=="bilinear" } {
        set T 1.0
        return [expr { [dflt atan {$val * $T/2.0}] * 2.0 / $T } ]
    }
    return $val
}

# Given a warping function T and its inverse T', we wish to map the
# interval [cs-ps,cs+ps] <--> [cz-pz,cz+pz].  Solving:
#     T(cs-ps) = cz-pz and T(cs+ps) = cz+pz
# prewarp:  cs = (T'(cz+pz) + T'(cz-pz))/2 and ps = (T'(cz+pz) - T'(cz-pz))/2
# postwarp: cz = ( T(cs+ps) +  T(cs-ps))/2 and pz = ( T(cs+ps) -  T(cs-ps))/2
#
proc xpSFD.WarpCP { dlg dir c_var p_var } {
    upvar $c_var c $p_var p
    set orig_c $c; set orig_p $p
    set x1 [xpSFD.${dir}WarpVal $dlg [expr { $c+$p }]]
    set x2 [xpSFD.${dir}WarpVal $dlg [expr { $c-$p }]]
    set c [expr { ($x1+$x2)/2.0 }]
    set p [expr { ($x1-$x2)/2.0 }]
#puts stdout "xpSFD.WarpCP: $orig_c $orig_p : $x1 $x2 : $c $p"
}

#
# Warping is a major mess when combined with spectral transformations.
# Because its non-linear, it works best when the spectral xform is done
# in z-domain.  Then we can do prewarping around baseband.
#
proc xpSFD.UpdateParams { dlg } {
    global errorInfo
    set func xpSFD.Update
    upvar #0 $dlg fvars $dlg.sx sx $dlg.sz sz

    xpSFD.GetParams $dlg

    set sx(active) 0
    set sz(active) 0

    set val_pass_s   $fvars(val-w_pass)
    set val_pass_z   $fvars(val-w_pass)
    if { "$fvars(pass)"=="lp" || "$fvars(pass)"=="hp" } {
	set val_center_s 0; set val_center_z 0
    } else {
        set val_center_s $fvars(val-w_center)
        set val_center_z $fvars(val-w_center)
    }

    #
    # The bilinear mapping s-->z causes a frequency distortion.
    # Thus the centers and widths in s & z domains may be different.
    # Note that this affect is independent of any prewarping.
    #
    set warpB [expr { "$fvars(domain)"=="z" && "$fvars(mapping)"=="bilinear" } ]
    if { $warpB } {
	if { "$fvars(passdomain)"=="z" } {
	    # bilinear sees low-pass only
	    if { $fvars(prewarpB) } {
		set val_pass_s [xpSFD.PreWarpVal $dlg $val_pass_s]
	    } else {
		set val_pass_z [xpSFD.PostWarpVal $dlg $val_pass_z]
	    }
	} else {
	    # bilinear sees band-pass
	    if { $fvars(prewarpB) } {
		xpSFD.WarpCP $dlg Pre val_center_s val_pass_s
	    } else {
		xpSFD.WarpCP $dlg Post val_center_z val_pass_z
	    }
	}
    }

    #
    # Create the "base" slp filter based on the design and params
    #
    set cmdargs ""
    foreach p $fvars(cmdparams) {
	if { "$p"=="w_pass" } {
	    lappend cmdargs $val_pass_s
	} else {
	    lappend cmdargs $fvars(val-$p)
	}
    }
    view clear $fvars(slp)
    if [catch {eval $fvars(cmd) $fvars(slp) $cmdargs} error] then {
	error "$func: $fvars(cmd) failed" $errorInfo
    }
    view event $fvars(slp)

    #
    # Configure the spectral transformation
    #
    if { "$fvars(pass)"!="lp" } {
	set val_pass   [set val_pass_$fvars(passdomain)]
	set val_center [set val_center_$fvars(passdomain)]
	set sx(Type) "$fvars(passdomain)lpto$fvars(pass)"
	set sx(OldCenter) 0
	set sx(OldWidth) $val_pass
	set sx(NewCenter) [expr {"$fvars(pass)"=="hp" ? 0 : $val_center}]
	set sx(NewWidth) $val_pass
    }
    #
    # Configure the mapping
    #
    if { "$fvars(mapping)"=="impulse" } {
	set sz(numzeros) $fvars(val-n_zeros)
    }

    foreach p $fvars(activate) {
#puts stdout "activate $p"
	if [catch {set ${p}(active) 1} error] then {
	    error "$func: $p failed" $errorInfo
	}
    }
}

proc xpSFD.SafeUpdateParams { dlg } {
    upvar #0 $dlg fvars
    global errorInfo

    if { $fvars(needUpdateB) } {
	return [xpSFD.SafeUpdate $dlg]
    }

    if [catch {xpSFD.UpdateParams $dlg} error] {
	xpStd.DlgErr $dlg xpSFD.SafeUpdateParams $error $errorInfo
	return 0
    }
    xpStd.DlgErrClear $dlg
    return 1
}

proc xpSFD.SafeUpdate { dlg } {
    upvar #0 $dlg fvars
    global errorInfo

    if [catch {xpSFD.UpdatePipes $dlg} error] {
	xpStd.DlgErr $dlg xpSFD.SafeUpdate $error $errorInfo
	return 0
    }
    set fvars(needUpdateB) 0
    return [xpSFD.SafeUpdateParams $dlg]
}

proc xpSFD.Ok { dlg } {
    if { ! [xpSFD.SafeUpdate $dlg] } {
	return
    }
    xpSFD.Popdown $dlg
}


proc xpSFD.Popdown { dlg } {
    focus none
    windowunmap $dlg
}

proc xpSFD.ShowAll { dlg } {
    upvar #0 $dlg fvars

    set fourierGraph [xpView.DisplayByClass $fvars(viewspec) CplxSeqFreq 1]
    set impulseGraph [xpView.DisplayByClass $fvars(viewspec) CplxSeqTime 1]
    foreach view $fvars(allviews) {
        xpView.DisplayByClass $view SilarRep 1

	foreach tinfo [list [list $fourierGraph CplxSeqFreq] [list $impulseGraph CplxSeqTime]] {
	    set vinfo [xpView.GetByClass $view [lindex $tinfo 1]]
	    set vspec [lindex $vinfo 0].[lindex $vinfo 1]
#puts stdout "Adding ``$view'' --> ``$vspec'' to ``[lindex $tinfo 0]''"
	    xpCSG.InsSeq [lindex $tinfo 0] $vspec "" 1
	}
    }
}

proc xpStdFltrDlg { dlg fltrname viewname } {
    upvar #0 $dlg fvars

    set viewspec $fltrname.$viewname
    set fltrinfo [filter info $fltrname]
    set fvars(viewspec)		$viewspec
    set fvars(domain)		[lindex $fltrinfo 1]
    set fvars(design)		butterworth
    set fvars(pass)		lp
    set fvars(passdomain)	z
    set fvars(mapping)		bilinear
    set fvars(prewarpB)		1
    set fvars(oldparams)	""
    set fvars(params)		""
    set fvars(clear)		1
    set fvars(autoupdate)	0
    set fvars(updateactive)	0
    set fvars(needUpdateB)	1
    set fvars(allviews)		""

    xpStdTitleBar $dlg.topbar "Standard Filter Design" "(xpole)Std Fltr Design"
    pack append $dlg $dlg.topbar { top fillx }
    menu $dlg.topbar.title.m
    label $dlg.name -text "[lindex $fltrinfo 2].$viewname"
    pack append $dlg $dlg.name { top }
    

    frame $dlg.mid -bd 2; pack append $dlg $dlg.mid { top fillx frame n }

    set dd $dlg.mid.design; set v ${dlg}(design)
    frame $dd -relief sunken -bd 2; pack append $dlg.mid $dd { left filly }
    pack append $dd [label $dd.title -text Design] top
    pack append $dd [radiobutton $dd.butterworth -text "Butterworth" -var $v \
      -rel flat -anc w] { top fill}
    pack append $dd [radiobutton $dd.chebyshev1 -text "Chebyshev-I" -var $v \
      -rel flat -anc w] { top fill}
    pack append $dd [radiobutton $dd.chebyshev2 -text "Chebyshev-II" -var $v \
      -rel flat -anc w] { top fill}
    pack append $dd [radiobutton $dd.elliptical -text "Elliptical" -var $v \
      -rel flat -anc w] { top fill}

    if { "$fvars(domain)"=="z" } {
	set ddm $dlg.mid.map
	pack append $dlg.mid [frame $ddm -relief sunken -bd 2] { left filly }

	set dd $ddm.map
	pack append $ddm [frame $dd] { top expand fill }
	pack append $dd [label $dd.title -text "Mapping"] { top fillx }
	foreach p { {bilinear "Bilinear"} {matchedz "Matched Z"} 
	  {impulse "Impulse Inv." } } {
	    pack append $dd [radiobutton $dd.[lindex $p 0] -text [lindex $p 1] \
	      -var ${dlg}(mapping) -rel flat -anc w] { top fill }
	}
        
	# XXX: the button padx option is broken in tk2.3
	set dd $ddm
	pack append $dd [checkbutton $dd.prewarp -text "Prewarp" \
	  -var ${dlg}(prewarpB) -rel flat -padx 10 -anc e] { top frame w }

    }

    set dd $dlg.mid.pass
    frame $dd -relief sunken -bd 2; pack append $dlg.mid $dd { left filly }
    pack append $dd [label $dd.title -text Band] top
    foreach p { {lp "Lo Pass"} {hp "Hi Pass"} {bp "Band Pass"} \
      {bs "Band Stop"} } {
	set butname [lindex $p 0]
	set buttext [lindex $p 1]
        pack append $dd [radiobutton $dd.$butname -text $buttext \
	  -var ${dlg}(pass) -rel flat -anc w] { top fill }
    }
    if { "$fvars(domain)"=="z" } {
	set ddp $dlg.mid.pass
	set dd $ddp.banddomain
	pack append $ddp [frame $dd] { top expand fill }
	pack append $dd [label $dd.title -text "Band Domain"] { top fillx }
	foreach p { {s "S"} {z "Z"} } {
	    pack append $dd [radiobutton $dd.[lindex $p 0] -text [lindex $p 1] \
	      -var ${dlg}(passdomain) -rel flat -anc w] { top fill }
	}
    }

    set dd $dlg.mid.vals
    frame $dd -relief sunken -bd 2; pack append $dlg.mid $dd { left expand fill}
    pack append $dd [label $dd.title -text Parameters] top
    # name	label intB	default lo hi	min max
    foreach p { \
      {order	N 1		3 1 10		1 ""} \
      {w_pass	w_pass 0	.25 .1 3.0	0 ""} \
      {w_tr	w_tr 0		.1 .1 3.0	0 ""} \
      {delta	delta 0		.1 0.0 1.0	0 1} \
      {w_center	w_ctr 0		.5 .1 3.0	0 ""} \
      {n_zeros	n_zeros 1	-1 -1 10	-1 ""} \
      } {
	set valname [lindex $p 0]
	set vallbl  [lindex $p 1]
	set valdef  [lindex $p 3]
	frame $dd.$valname
	pack append $dd.$valname [label $dd.$valname.lbl -text $vallbl \
	  -rel flat -anc w -width 7] { left }
	pack append $dd.$valname [xpscale $dd.$valname.val \
	  -wval $valdef -wlo [lindex $p 4] -whi [lindex $p 5] \
	  -wmin [lindex $p 6] -wmax [lindex $p 7] -integer [lindex $p 2] \
	  -editlimits 0 -orient horiz \
	  -command "xpSFD.ValCB $dlg $valname"] { left fillx expand }

	set fvars(val-$valname) $valdef
        trace var fvars(val-$valname) w "xpSFD.TraceVals $dlg"
    }

    pack append $dlg [xpStd.DlgErrCreate $dlg $dlg.msg] { top fillx }

    set dd $dlg.opts
    frame $dd -bd 2; pack append $dlg $dd { top fill }
    label $dd.lbl -text "Options: "
    checkbutton $dd.clear -text "Clear" -var ${dlg}(clear)
    pack append $dd $dd.clear { left expand fill }
    checkbutton $dd.update -text "AutoUpdate" -var ${dlg}(autoupdate)
    pack append $dd $dd.update { left expand fill }

    set dd $dlg.buts
    pack append $dlg [frame $dd -bd 2] {top fill}
    pack append $dd [button $dd.ok -text "Ok" \
      -com "xpSFD.Ok $dlg"] {left expand fill}
    pack append $dd [button $dd.update -text "Update" \
      -com "xpSFD.SafeUpdate $dlg"] {left expand fill}
    pack append $dd [button $dd.showall -text "Show All" \
      -com "xpSFD.ShowAll $dlg"] {left expand fill}
    pack append $dd [button $dd.cancel -text "Cancel" \
      -com "xpSFD.Popdown $dlg"] {left expand fill}


    trace var fvars(design) w "xpSFD.TraceDesign $dlg"
    trace var fvars(mapping) w "xpSFD.TraceDesign $dlg"
    trace var fvars(passdomain) w "xpSFD.TraceDesign $dlg"
    trace var fvars(pass) w "xpSFD.TraceDesign $dlg"
    trace var fvars(params) w "xpSFD.TraceParams $dlg"
    trace var fvars(prewarpB) w "xpSFD.TraceVals $dlg"

    trace var fvars(autoupdate) w "xpSFD.TraceUpdate $dlg"

    # Trigger a recalc
    set fvars(design) butterworth
}

proc xpStdFltrDlgPopup { {fltrview ""} } {
    if { "$fltrview"=="" } {
	xpViewSelDlg "Standard Filter Dialog" "xpStdFltrDlgPopup"
	return
    }
    set viewinfo [xpView.GetByClass $fltrview SilarRep]
    set fltrname [lindex $viewinfo 0]
    set viewname [lindex $viewinfo 1]
    set dlg .stdfltr-$fltrname-$viewname
    set viewspec $fltrname.$viewname

    if ![topWin.ExistB $dlg] {
        toplevel $dlg
	wm minsize $dlg 0 0
	wm title $dlg "XPole StdFltr $viewspec" 
	wm iconname $dlg "XPole StdFltr $viewspec" 
	xpStdFltrDlg $dlg $fltrname $viewname
    } else {
        windowmap $dlg
    }
}
